/*
 * Copyright 2016 - 2023 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <nostalgia/core/gfx.hpp>
#include <nostalgia/foundation/media.hpp>

#include "gfx.hpp"

namespace nostalgia::core {

ox::Error initConsole(Context *ctx) noexcept {
	constexpr ox::FileAddress TilesheetAddr("/TileSheets/Charset.ng");
	constexpr ox::FileAddress PaletteAddr("/Palettes/Charset.npal");
	setBgStatus(ctx, 0b0001);
	setBgCbb(ctx, 0, 0);
	return loadBgTileSheet(ctx, 0, TilesheetAddr, PaletteAddr);
}

struct TileSheetData {
	ox::Vector<uint32_t> pixels;
	int width = 0;
	int height = 0;
};

ox::Result<TileSheetData> loadTileSheet(Context *ctx, const CompactTileSheet &tilesheet) noexcept {
	const unsigned bytesPerTile = tilesheet.bpp == 8 ? PixelsPerTile : PixelsPerTile / 2;
	const auto tiles = tilesheet.pixels.size() / bytesPerTile;
	constexpr int width = 8;
	const int height = 8 * static_cast<int>(tiles);
	ox::Vector<uint32_t> pixels;
	if (bytesPerTile == 64) { // 8 BPP
		pixels.resize(tilesheet.pixels.size());
		for (std::size_t i = 0; i < tilesheet.pixels.size(); ++i) {
			pixels[i] = tilesheet.pixels[i];
		}
	} else { // 4 BPP
		pixels.resize(tilesheet.pixels.size() * 2);
		for (std::size_t i = 0; i < tilesheet.pixels.size(); ++i) {
			pixels[i * 2 + 0] = tilesheet.pixels[i] & 0xF;
			pixels[i * 2 + 1] = tilesheet.pixels[i] >> 4;
		}
	}
	const auto rd = ctx->rendererData<void>();
	renderer::loadSpriteTexture(rd, pixels.data(), width, height);
	return TileSheetData{std::move(pixels), width, height};
}

ox::Error loadBgTileSheet(Context *ctx,
                          unsigned cbb,
                          const ox::FileAddress &tilesheetAddr,
                          const ox::FileAddress &paletteAddr) noexcept {
	oxRequire(tilesheet, readObj<CompactTileSheet>(ctx, tilesheetAddr));
	oxRequire(palette, readObj<Palette>(ctx, paletteAddr ? paletteAddr : tilesheet->defaultPalette));
	oxRequire(tsd, loadTileSheet(ctx, *tilesheet));
	const auto rd = ctx->rendererData<void>();
	renderer::loadBgTexture(rd, cbb, tsd.pixels.data(), tsd.width, tsd.height);
	renderer::loadBgPalette(rd, *palette);
	return {};
}

ox::Error loadSpriteTileSheet(Context *ctx,
                              const ox::FileAddress &tilesheetAddr,
                              const ox::FileAddress &paletteAddr) noexcept {
	oxRequire(tilesheet, readObj<CompactTileSheet>(ctx, tilesheetAddr));
	oxRequire(palette, readObj<Palette>(ctx, paletteAddr ? paletteAddr : tilesheet->defaultPalette));
	oxRequire(tsd, loadTileSheet(ctx, *tilesheet));
	const auto rd = ctx->rendererData<void>();
	renderer::loadSpriteTexture(rd, tsd.pixels.data(), tsd.width, tsd.height);
	renderer::loadSpritePalette(rd, *palette);
	return {};
}

void puts(Context *ctx, int column, int row, ox::CRStringView str) noexcept {
	const auto col = static_cast<unsigned>(column);
	for (auto i = 0u; i < str.bytes(); ++i) {
		setTile(ctx, 0, static_cast<int>(col + i), row, static_cast<uint8_t>(charMap[static_cast<uint8_t>(str[i])]));
	}
}

}
